ko.collaboration.friends = {};

(function() {
    this.update = function() {
        this.updateFriendsList();
    };

    this.log = ko.logging.getLogger("ko.collaboration.friends");

    /**
     * List of all available friends. Gets initialized by updateFriendsList()
     * once friends are fetched from the server for the first time.
     */
    this.allFriends = null;

    /**
     * Prompts the user to supply a username and makes an API call to create a
     * friend request for that user.
     */
    this.sendFriendRequest = function sendFriendRequest() {
        var username = ko.dialogs.prompt("Please enter the email address of the contact you would like to add", "Email address",
                                         null, "Collaboration");
        if (username) {
            var this_ = this;
            let errorCallback = function(req) {
                if (req.status == 404) {
                    ko.dialogs.alert("User not found",
                                     "The email address you supplied does not belong to a known Komodo Collaboration user.");
                } else {
                    this_._displayError();
                }
            };
            ko.collaboration.api.addFriendAsync(username,
                                             ko.collaboration.utils.bind(this.update, this),
                                             errorCallback);
        }
    };

    /**
     * Fetches usernames of all friends and updates the friends list. When that
     * is done, continue with friend requests.
     */
    this.updateFriendsList = function updateFriendsList() {
        ko.collaboration.api.getFriendsAsync(
            this._handleFriendsListResponse.bind(this),
            this._displayError.bind(this));
    };

    this._handleFriendsListResponse = function _handleFriendsListResponse(response) {
        if (!response.hasOwnProperty('friends')) {
            this.log.error("Invalid response data from server: " +
                      JSON.stringify(response));
            this._displayError();
        }
        this._clearFriendsLists();
        this.allFriends = {};
        for each (let [userId, userName] in Iterator(response.friends)) {
            this.allFriends[userId] = userName;
        }
        for each (let [userId, userName] in this.getAllFriendsSorted()) {
            this._addFriend(userId, userName);
        }
        this._updateFriendRequests();
    };

    /**
     * Converts the unsorted object structure `allFriends` into an array of
     * arrays sorted by username in ascending order, and returns it.
     */
    this.getAllFriendsSorted = function getAllFriendsSorted() {
        var sorted = [];
        for each (let [userId, userName] in Iterator(this.allFriends)) {
            sorted.push([userId, userName]);
        }
        sorted.sort(function(a, b) {
            return a[1] < b[1] ? -1 : 1; // Sort by userName
        });
        return sorted;
    };

    this._clearFriendsLists = function _clearFriendsLists() {
        let list = document.getElementById('friends-list');
        if (list) {
            while (list.itemCount > 0) {
                list.removeItemAt(0);
            }
        }
    };

    this.friendRequests = {};

    /**
     * Get all friend requests from the server, and add those to the contacts
     * list. Should only be called from updateFriends() to make sure that the
     * list is in a clean state.
     */
    this._updateFriendRequests = function _updateFriendRequests() {
        let successCallback = ko.collaboration.utils.bind(function(response) {
            if (!response.hasOwnProperty('requests')) {
                this.log.error("Invalid response data from server: " +
                          JSON.stringify(response));
                this._displayError();
            }
            for each (let [userId, userName] in Iterator(response.requests)) {
                this._addFriendRequest(userId, userName);
            }
        }, this);
        let errorCallback = ko.collaboration.utils.bind(this._displayError, this);

        ko.collaboration.api.getFriendRequestsAsync(successCallback,
                                                    errorCallback);
    };

    /**
     * Adds a contact to the contact list.
     */
    this._addFriend = function _addFriend(userId, userName) {
        var friendsList = document.getElementById('friends-list');
        var item = document.createElement('richlistitem');
        item.setAttribute('class', 'friends-list-friend');
        // Insert the node into the DOM before using the xbl magic.
        friendsList.appendChild(item);
        item.userId = userId;
        item.userName = userName;
    };

    /**
     * Adds a friend request entry to the contacts list.
     */
    this._addFriendRequest = function _addFriendRequest(userid, username) {
        var friendsList = document.getElementById('friends-list');
        var item = document.createElement('richlistitem');
        item.setAttribute('class', 'friends-list-request');
        // Insert the node into the DOM before using the xbl magic.
        friendsList.appendChild(item);
        item.userId = userid;
        item.userName = username;
    };

    this._displayError = function _displayError() {
        this._clearFriendsLists();
        var friendsList = document.getElementById('friends-list');
        var item = document.createElement('richlistitem');
        item.setAttribute('class', 'friends-list-message');
        // Insert the node into the DOM before using the xbl magic.
        friendsList.appendChild(item);
        item.setAttribute('description', 'An error occured while trying to ' +
                          'get your friends list. Please click the reload ' +
                          'button to try again.');
    };
}).apply(ko.collaboration.friends);
