/* Copyright (c) 2003-2008 ActiveState Software Inc.
   See the file LICENSE.txt for licensing information. */

/* Source Code Control commit window.
 *
 * Usage:
 *  All dialog interaction is done via an object passed in and out as the first
 *  window argument: window.arguments[0].
 *      .sccSvc         The SCC component the files belong to.
 *      .callbackFn     Callback handler for when the window is closed.
 *      .type           The type of command the dialog is used for, can be one
 *                      of: "commit", "add", "delete", "edit", "revert",
 *                      "changelist".
 *      .urls           (optional) The file urls that are going to be committed.
 *                      The URLs are automatically converted for display via
 *                      ko.uriparse.displayPath() and will initially be the
 *                      only files marked as "checked" in the tree.
 *      .message        (optional) the default check-in message
 *      .title          (optional) the title to be set on the window.
 *      .changelistName (optional) the name of the changelist
 *
 *  On return window.arguments[0] has:
 *      .retval         "OK" or "Cancel"
 *      .urls           a list of selected urls (only if .retval=="OK")
 *      .message        the user's checkin message
 *      .changelistName (only for type='changelist') the name of the changelist
 *
 * Dev Notes: How this dialog works.
 *
 * The dialog will initially display the passed in files with the last known
 * Komodo scc status for these files. An asynchronous scc update check will be
 * launched to find out the latest scc status.
 * 
 * 1) If only one file uri is passed in, the asynchronous status check will be
 *    made on the file's directory, loading all additional found items into the
 *    tree.
 * 2) When more than one file uri is passed in, the asynchronous scc status
 *    check will only be performed on these files themselves, no additional
 *    modified file items are loaded into the tree.
 */

var log = ko.logging.getLogger("scc.commit");
//log.setLevel(ko.logging.LOG_DEBUG);

    // Localization.
var _bundle = Components.classes["@mozilla.org/intl/stringbundle;1"]
                .getService(Components.interfaces.nsIStringBundleService)
                .createBundle("chrome://komodo/locale/scc.properties");
    // Type of scc action that will be performed.
var gType = "commit";
    // List of URLs from which the user can select.
var gURLs = [];
    // Common base uri for all files in the gSCCItems list.
var gBaseUri = null;
    // List of all koSCCxxxItems objects. Will be used by the tree view when
    // toggling between the "show all/checked" options. All updates made in
    // loading/adding/removing of items should be made to this list and then
    // propogated to the tree view!
var gSCCItems = [];
    // List of items that are currently asynchronously loading...
var gLoadingItems = [];
    /**
     * SCC service instance, passed in to the dialog.
     * @type Components.interfaces.koISCC
     */
var gSCCSvc = null;
    // Tree view instance (CommitTreeView).
var gURLTreeView = null;


/**
 * Custom koSCCxxxItem classes, used to populate the tree view.
 */
function koSCCBaseItem(uri, relPath, status, outOfSync, checked) {
    this.uriSpec = uri;
    this.relativePath = relPath;
    this.status = status;
    this.isOutOfSync = outOfSync;
    this.isChecked = checked;
};
koSCCBaseItem.prototype.copyItemFileStatus = function(item) {
    this.status = item.status;
    this.isOutOfSync = item.isOutOfSync;
};
koSCCBaseItem.prototype.getStatusAsString = function() {
    switch (this.status) {
        case Components.interfaces.koISCC.STATUS_UNKNOWN:
            return "SCC status is unknown";
        case Components.interfaces.koISCC.STATUS_OK:
            return "File is unmodified";
        case Components.interfaces.koISCC.STATUS_MODIFIED:
            return "File is modified";
        case Components.interfaces.koISCC.STATUS_ADDED:
            return "File has been added";
        case Components.interfaces.koISCC.STATUS_DELETED:
            return "File is marked as deleted";
        case Components.interfaces.koISCC.STATUS_REPLACED:
            return "File has been replaced";
        case Components.interfaces.koISCC.STATUS_MODIFIED_PROPERTY:
            return "Modified property";
        case Components.interfaces.koISCC.STATUS_CONFLICT:
            return "SCC conflict";
    }
    return "";
}


// Files that were passed into the commit dialog via the .urls object. This
// will usually mean the minimum set of files the user wants to commit. These
// items *might* not have any known scc status.
function koSCCFileItem(koFile) {
    var uri = koFile.URI;
    var relPath = uri.substring(gBaseUri.length + 1);
    var status = getStatusFromkoIFile(koFile);
    //dump("status: " + status + ", uri: " + koFile.URI + "\n");
    koSCCBaseItem.apply(this, [uri, relPath, status, koFile.sccNeedSync, true]);
    this.isDirectory = koFile.isDirectory;
};
koSCCFileItem.prototype = new koSCCBaseItem();
koSCCFileItem.prototype.contructor = koSCCFileItem;


// Files that were retrieved from the scc status calls.
function koSCCFileStatusItem(item) {
    koSCCBaseItem.apply(this, [item.uriSpec, item.relativePath, item.status, item.isOutOfSync, false]);
    var fileSvc = Components.classes["@activestate.com/koFileService;1"].
                    getService(Components.interfaces.koIFileService);
    var koFile = fileSvc.getFileFromURI(item.uriSpec);
    this.isDirectory = koFile && koFile.isDirectory || false;
};
koSCCFileStatusItem.prototype = new koSCCBaseItem();
koSCCFileStatusItem.prototype.contructor = koSCCFileStatusItem;


// Item to show that the scc status is being asynchronously loaded.
function koSCCLoadingItem(message) {
    var strbundle = document.getElementById("scc_strings");
    message = strbundle.getFormattedString("loadingMessage", [ gSCCSvc.name, message ]);
    koSCCBaseItem.apply(this, ["", message, Components.interfaces.koISCC.STATUS_OK, false, false]);
    this.aborted = false;
};
koSCCLoadingItem.prototype = new koSCCBaseItem();
koSCCLoadingItem.prototype.contructor = koSCCLoadingItem;

// Item to show that an error in the tree.
function koSCCErrorItem(message) {
    koSCCBaseItem.apply(this, ["", message, Components.interfaces.koISCC.STATUS_OK, false, false]);
};
koSCCErrorItem.prototype = new koSCCBaseItem();
koSCCErrorItem.prototype.contructor = koSCCErrorItem;



/**
 * TreeView for the commit dialog.
 */
function CommitTreeView(initial_rows) {
        // Call the parent initializer.
    xtk.dataTreeView.apply(this, [initial_rows]);
        // Only show the items marked as "checked" in the tree.
    this.onlyShowCheckedItems = false;
        // Callback handler for on modify notifications.
    this.onModifiedCallback = null;
        // Atom service is used to set the tree cell css properties.
    this._atomService = Components.classes["@mozilla.org/atom-service;1"].
                            getService(Components.interfaces.nsIAtomService);
};
CommitTreeView.prototype = new xtk.dataTreeView();
CommitTreeView.prototype.contructor = CommitTreeView;
CommitTreeView.prototype.getCellText = function(row, column)
{
    if (column.id == "urls-column-name") {
        return this.rows[row].relativePath;
    }
    return "";
};
CommitTreeView.prototype.isEditable = function(row, column)
{
    return column.id == "urls-column-checkbox";
};
CommitTreeView.prototype.getCellValue = function(row, column)
{
    if (column.id == "urls-column-checkbox") {
        return this.rows[row].isChecked;
    }
    return null;
};
CommitTreeView.prototype.setCellValue = function(row, column, value)
{
    if (column.id == "urls-column-checkbox") {
        if (value.toLowerCase() == "false") {
            this.rows[row].isChecked = false;
        } else {
            this.rows[row].isChecked = true;
        }
        this.tree.invalidateRow(row);
        if (this.onModifiedCallback) {
            this.onModifiedCallback();
        }
    }
};
CommitTreeView.prototype.getCellProperties = function(row, column, properties)
{
    var item = this.rows[row];
    if (item instanceof koSCCLoadingItem) {
        properties.AppendElement(this._atomService.getAtom("sccLoading"));
    } else if (item instanceof koSCCErrorItem) {
        properties.AppendElement(this._atomService.getAtom("sccError"));
    }
    if (column.id == "urls-column-name") {
        // Set the scc icon for the status column.
        properties.AppendElement(this._atomService.getAtom("nameColumn"));
        var status = item.status;
        if (status == Components.interfaces.koISCC.STATUS_UNKNOWN) {
            properties.AppendElement(this._atomService.getAtom("sccUnknown"));
        } else if (status == Components.interfaces.koISCC.STATUS_OK) {
            properties.AppendElement(this._atomService.getAtom("sccOk"));
        } else if (status == Components.interfaces.koISCC.STATUS_MODIFIED) {
            properties.AppendElement(this._atomService.getAtom("sccModified"));
        } else if (status == Components.interfaces.koISCC.STATUS_ADDED) {
            properties.AppendElement(this._atomService.getAtom("sccAdd"));
        } else if (status == Components.interfaces.koISCC.STATUS_DELETED) {
            properties.AppendElement(this._atomService.getAtom("sccDelete"));
        } else if (status == Components.interfaces.koISCC.STATUS_REPLACED) {
            properties.AppendElement(this._atomService.getAtom("sccReplaced"));
        } else if (status == Components.interfaces.koISCC.STATUS_MODIFIED_PROPERTY) {
            properties.AppendElement(this._atomService.getAtom("sccModifiedProperty"));
        } else if (status == Components.interfaces.koISCC.STATUS_CONFLICT) {
            properties.AppendElement(this._atomService.getAtom("sccConflict"));
        }

        // Add any out of sync details.
        if (item.isOutOfSync) {
            properties.AppendElement(this._atomService.getAtom("sccSync"));
        }
    }
};

CommitTreeView.prototype.getImageSrc = function(row, column)
{
    if (column.id == "urls-column-name") {
        var item = this.rows[row];
        if (!(item instanceof koSCCLoadingItem) &&
            !(item instanceof koSCCErrorItem)) {
            if (item.isDirectory) {
                // Set the native folder icon.

                return "moz-icon://stock/gtk-directory?size=16";



            } else {
                // Set the native file icon.
                return "moz-icon://" + item.relativePath + "?size=16";
            }
        }
    }
    return null;
};

CommitTreeView.prototype.setTreeItems = function(sccItems)
{
    if (this.onlyShowCheckedItems) {
        // Filter down the list of items.
        sccItems = sccItems.filter(function(item, index, array) { return item.isChecked });
    } else {
        // Make a copy of the list.
        sccItems = sccItems.concat();
    }
    this.rows = sccItems;
};

/**
 * Get the list of commitable urls for the supplied directory.
 * @param {array} urls  The list of urls to check the status for.
 * @param {boolean} recursive  Whether to perform the status check recursively.
 */
CommitTreeView.prototype.GetCommitableFiles = function(urls, recursive) {
    try {
        if (urls.length == 0) {
            // Nothing to do...
            return;
        }
        if (!gSCCSvc) {
            // Determine the scc type from the URIs provided.
            for (var url_index=0; url_index < urls.length; url_index++) {
                gSCCSvc = ko.scc.getServiceForUrl(urls[url_index]);
                if (gSCCSvc) {
                    break;
                }
            }
            if (!gSCCSvc) {
                ko.dialogs.alert("There is no known SCC type from the files provided");
                return;
            }
        }
        var loadingItem = null;
        var self = this;
        var async_callback = {
            "callback": function(result, data) {
                log.info("GetCommitableFiles:: callback result: " + result);
                var idx = gSCCItems.indexOf(loadingItem);
                if (idx >= 0) {
                    // Remove the loading item.
                    gSCCItems.splice(idx, 1);
                    log.debug("GetCommitableFiles:: removing the loading item");
                }
                try {
                    if (loadingItem.aborted) {
                        log.debug("GetCommitableFiles:: loading item was aborted");
                        return;
                    }
                    // "data" will be an array of koISCCFileStatusItem, update
                    // the tree view with this information.
                    if (result == Components.interfaces.koIAsyncCallback.RESULT_SUCCESSFUL) {
                        // Update any already existing items in the tree with
                        // the new status we have received.
                        var filterFn = function(item, index, array) {
                            var rowItem = _findSccItemWithUrl(gSCCItems, item.uriSpec);
                            if (rowItem != null) {
                                // Copy across what we know.
                                //dump("Found a match: " + item.uriSpec + "\n");
                                log.debug("GetCommitableFiles:: returned item already in the tree, updating: " +
                                          rowItem.uriSpec);
                                rowItem.copyItemFileStatus(item);
                                return false;
                            }
                            return true;
                        };
                        data = data.filter(filterFn);
                        // Turn all returned items into a koSCCFileStatusItem.
                        var mapperFn = function(item) { return new koSCCFileStatusItem(item) };
                        data = data.map(mapperFn);
                        gSCCItems = gSCCItems.concat(data);
                        rebaseSccItemUris(gBaseUri);
                    } else if (result == Components.interfaces.koIAsyncCallback.RESULT_ERROR) {
                        // data should contain the error message then.
                        if (!data) {
                            data = "Asynchronous GetCommitableFiles() failed.";
                        }
                        gSCCItems.push(new koSCCErrorItem(String(data)));
                        log.info("GetCommitableFiles:: error returned: " + data);
                    }

                    if (self.onModifiedCallback) {
                        self.onModifiedCallback();
                    }

                } catch (e) {
                    //dump("Exception: " + e + "\n");
                    gSCCItems.push(new koSCCErrorItem(e));
                    log.exception(e, "getCommitableFiles async callback failed");
                } finally {
                    self.setTreeItems(gSCCItems);
                    self = null;
                    // Ensure the tree cells are properly painted, as sometimes
                    // a glitch can occur where the cells don't have proper
                    // status images (on Linux).
                    window.setTimeout(function() { gURLTreeView.tree.invalidate() }, 200);
                }
            }
        }

        gSCCSvc.status(urls.length, urls, recursive, '', async_callback);

        var baseUri = gBaseUri;
        if (baseUri[baseUri.length - 1] != '/') {
            baseUri += '/';
        }
        var urlRelativizer = function(uri) {
            if (uri.length <= baseUri.length) {
                return ko.uriparse.baseName(uri);
            }
            return uri.substring(baseUri.length);
        };
        var relativePaths = urls.map(urlRelativizer);
        loadingItem = new koSCCLoadingItem(relativePaths.join(", "));
        gSCCItems.push(loadingItem);
        this.setTreeItems(gSCCItems);

    } catch(ex) {
        log.exception(ex, "Could not retrieve scc status for files " + urls);
    }
};


function CommitTreeViewController() {
}
// The following two lines ensure proper inheritance (see Flanagan, p. 144).
CommitTreeViewController.prototype = new xtk.Controller();
CommitTreeViewController.prototype.constructor = CommitTreeViewController;

CommitTreeViewController.prototype.is_cmd_selectAll_enabled = true;
CommitTreeViewController.prototype.do_cmd_selectAll = function() {
    gURLTreeView.selection.selectAll();
}



//---- interface routines for XUL

function OnLoad()
{
    try {
        var dialog = document.getElementById("window-commit");
        var strbundle = document.getElementById("scc_strings");

        // .type
        var type = window.arguments[0].type;
        if (typeof(type) != "undefined" && type != null) {
            gType = type;
            if (gType == 'changelist') {
                var changelist_vbox = document.getElementById("changelist_name_vbox");
                changelist_vbox.setAttribute("collapsed", "false");
                var changelistName = window.arguments[0].changelistName;
                if (changelistName) {
                    var changelist_textbox = document.getElementById("changelist_name_textbox");
                    changelist_textbox.value = changelistName;
                }
                var commitButton = document.getElementById("button_commit");
                commitButton.label = strbundle.getString("changelistButtonSaveLabel");
                commitButton.setAttribute("accesskey", strbundle.getString("changelistButtonSaveAccessKey"));
            }
        }

        // .sccSvc
        gSCCSvc = window.arguments[0].sccSvc;
        var scc_name = "<No SCC>";
        if (typeof(gSCCSvc) == "undefined" || gSCCSvc == null) {
            // We'll get the scc type when the user adds some files.
            gSCCSvc = null;
        } else {
            scc_name = gSCCSvc.name;
        }

        // .callbackFn
        var cbFn = window.arguments[0].callbackFn;
        if (!cbFn) {
            var msg = "Internal Error: no '.callbackFn' argument was provided. "
                      +"SCC Commit dialog: '"+gURLs+"'.";
            log.error(msg);
            alert(msg);
            window.close();
            return;
        }

        // .urls
        if (window.arguments[0].urls) {
            gURLs = window.arguments[0].urls.sort();
        }

        // .message
        var messageTextbox = document.getElementById("message");
        var message = window.arguments[0].message;
        if (typeof message != "undefined" && message != null) {
            messageTextbox.value = message;
        } else {
            messageTextbox.value = "";
        }

        // .title
        if (typeof window.arguments[0].title != "undefined" &&
            window.arguments[0].title != null) {
            document.title = window.arguments[0].title;
        } else {
            if (gType == 'changelist') {
                document.title = strbundle.getFormattedString("changelistWindowTitle", [ scc_name ]);
            } else {
                document.title = strbundle.getFormattedString("commitWindowTitle", [ scc_name ]);
            }
        }

        var urlsTree = document.getElementById("urls");
        gBaseUri = determineBaseUriFromUrls(gURLs);
        log.debug("gBaseUri: " + gBaseUri);
        gURLTreeView = new CommitTreeView(gSCCItems);
        gURLTreeView.onModifiedCallback = updateCommitButtonEnabledState;
        addChosenSccFiles(gURLs);
        updateSccFileStatus(gURLs);
        urlsTree.treeBoxObject.view = gURLTreeView;

        // Controller used to handle "selectAll" in the tree view.
        urlsTree.controllers.insertControllerAt(0, new CommitTreeViewController());

        messageTextbox.focus();
        messageTextbox.setSelectionRange(0, messageTextbox.textLength);

        updateCommitButtonEnabledState();

        // Size the xul window to match it's contents...
        window.sizeToContent();
    } catch(ex) {
        log.exception(ex);
    }
}


function updateVisibleTreeItems() {
    try {
        var show_checked_cbox = document.getElementById("only_show_checked_cbox");
        gURLTreeView.onlyShowCheckedItems = show_checked_cbox.checked;
        gURLTreeView.setTreeItems(gSCCItems);
    } catch(ex) {
        log.exception(ex);
    }
}

/**
 * Return the koISCC status for the given koIFileEx object.
 * @param file {Components.interfaces.koIFileEx}
 */
function getStatusFromkoIFile(file) {
    var status = Components.interfaces.koISCC.STATUS_UNKNOWN;
    try {
        var action = file.sccAction;
        if (file.sccConflict || (action == 'conflict')) {
            status = Components.interfaces.koISCC.STATUS_CONFLICT;
        } else if (action == 'integrate' || action == 'edit') {
            status = Components.interfaces.koISCC.STATUS_MODIFIED;
        } else if (action == 'add') {
            status = Components.interfaces.koISCC.STATUS_ADDED;
        } else if (action == 'delete') {
            status = Components.interfaces.koISCC.STATUS_DELETED;
        } else if (action == 'ok' || (!action && file.sccType)) {
            status = Components.interfaces.koISCC.STATUS_OK;
        }
    } catch(ex) {
        log.exception(ex);
    }
    return status;
}

function determineBaseUriFromUrls(urls) {
    var baseUri;
    var koOsPathSvc = Components.classes["@activestate.com/koOsPath;1"].
                        getService(Components.interfaces.koIOsPath);
    if (urls.length == 0) {
        return "";
    } else if (urls.length == 1) {
        // If it's a file path, load the status from the directory instead.
        if (koOsPathSvc.isfile(ko.uriparse.URIToLocalPath(urls[0]))) {
            // Not using ko.uriparse.dirname because it removes the URI section.
            baseUri = urls[0].match("^(.*)/(.*?)$")[1];
        } else {
            baseUri = urls[0];
        }
    } else {
        log.debug("urls: " + urls);
        baseUri = ko.uriparse.commonURIPrefixFromURIs(urls);
    }
    log.debug("determineBaseUriFromUrls: " + baseUri);

    var label = document.getElementById("commit_dialog_location_label_value");
    label.setAttribute("value", ko.uriparse.URIToLocalPath(baseUri));

    return baseUri;
}

function rebaseSccItemUris(baseUri) {
    var item;
    log.debug("rebasing uris to: " + baseUri);
    for (var i=0; i < gSCCItems.length; i++) {
        item = gSCCItems[i];
        if (!(item instanceof koSCCLoadingItem) &&
            !(item instanceof koSCCErrorItem)) {
            item.relativePath = item.uriSpec.substring(baseUri.length + 1);
            if (!item.relativePath) {
                // It's a directory that has been modified, we need to rebase
                // to the parent directory - bug 78026.
                baseUri = ko.uriparse.pathToURI(ko.uriparse.dirName(baseUri));
                rebaseSccItemUris(baseUri);
                return;
            }
        }
    }
    gBaseUri = baseUri;
    var basePath = ko.uriparse.URIToLocalPath(baseUri);
    var label = document.getElementById("commit_dialog_location_label_value");
    label.setAttribute("value", basePath);
}

/**
 * Add the files to the tree list and use the status of whatever we already know
 * about the file.
 */
function addChosenSccFiles(urls) {
    try {
        var i;
        var url;
        var status;
        var item;
        var koFile;
        var fileSvc = Components.classes["@activestate.com/koFileService;1"].
                        getService(Components.interfaces.koIFileService);
        var defaultItems = [];
        for (i=0; i < urls.length; i++) {
            url = urls[i];
            // Only add file URLs
            koFile = fileSvc.getFileFromURI(url);
            if (koFile && !koFile.isDirectory) {
                item = new koSCCFileItem(koFile);
                defaultItems.push(item);
            }
        }
        gSCCItems = gSCCItems.concat(defaultItems);
        gURLTreeView.setTreeItems(gSCCItems);
    } catch(ex) {
        log.exception(ex);
    }
}

function updateSccFileStatus(urls) {
    try {
        var koOsPathSvc = Components.classes["@activestate.com/koOsPath;1"].
                            getService(Components.interfaces.koIOsPath);
        var statusUrls = urls;
        // If it's a single file path, load the status from the directory
        // instead.
        if ((urls.length == 1) &&
            (koOsPathSvc.isfile(ko.uriparse.URIToLocalPath(urls[0])))) {
            statusUrls = [ko.uriparse.dirName(urls[0])];
        }

        var recursive = true;
        gURLTreeView.GetCommitableFiles(statusUrls, recursive);
    } catch(ex) {
        log.exception(ex);
    }
}

function _getSelectedTreeItems()
{
    var selected_items = [];
    var selection = gURLTreeView.selection;
    var num_ranges = selection.getRangeCount();
    for (var i = 0; i < num_ranges; ++i) {
        var min = {}, max = {};
        selection.getRangeAt(i, min, max);
        for (var j = min.value; j <= max.value; ++j) {
            selected_items.push(gURLTreeView.rows[j]);
        }
    }
    return selected_items;
}

function on_keypress(event)
{
    try {
        var urlsTree = document.getElementById("urls");
        // We only want to perform an action if there are some selected items in
        // the tree.
        if (urlsTree.currentIndex >= 0) {
            if (event.charCode == event.DOM_VK_SPACE) {
                var selection = gURLTreeView.selection;
                var num_ranges = selection.getRangeCount();
                for (var i = 0; i < num_ranges; ++i) {
                    var min = {}, max = {};
                    selection.getRangeAt(i, min, max);
                    for (var j = min.value; j <= max.value; ++j) {
                        gURLTreeView.rows[j].isChecked = !gURLTreeView.rows[j].isChecked;
                    }
                    gURLTreeView.tree.invalidateRange(min.value, max.value);
                }
                return false;
            }
        }
    } catch(ex) {
        log.exception(ex);
    }
    return true;
}

/**
 * Find the scc item in the item list provided that matches the given url.
 * Returns the item when found, otherwise returns null.
 *
 * Note: The url comparison is case-insensitive for Windows.
 * 
 * @returns koSCCBaseItem
 */
function _findSccItemWithUrl(sccItems, url) {



    var item_url;
    for (var i=0; i < sccItems.length; i++) {
        item_url = sccItems[i].uriSpec;



        if (url == item_url) {
            return sccItems[i];
        }
    }
    return null;
}

function _urlIsInSccItemsList(url) {
    return (_findSccItemWithUrl(gSCCItems, url) != null);
}

function _getCheckedItems() {
    return gSCCItems.filter(function(item, index, array) {
                            return item.isChecked;
                          });
}

function _getUrlsFromItems(items) {
    return items.map(function(item) {
                        return item.uriSpec;
                     });
}

function _getRelativePathsFromItems(items) {
    return items.map(function(item) {
                        return item.relativePath;
                     });
}

function _getCheckedUrls() {
    return _getUrlsFromItems(_getCheckedItems());
}

function _getCheckedRelativePaths() {
    return _getRelativePathsFromItems(_getCheckedItems());
}

function updateCommitButtonEnabledState()
{
    var dialog = document.getElementById("window-commit");
    var okButton = document.getElementById("button_commit");
    if (!gURLTreeView.rows.some(function(elem) { return elem.isChecked })) {
        okButton.setAttribute("disabled", "true");
    } else if (okButton.hasAttribute("disabled")) {
        okButton.removeAttribute("disabled");
    }
}

// --- Context menu handlers.

function DiffSelectedItems()
{
    var relpaths = _getRelativePathsFromItems(_getSelectedTreeItems());
    if (relpaths.length > 0) {
        ko.scc.DiffRelative(gSCCSvc, gBaseUri, relpaths);
    }
}

function RemoveSelectedItems()
{
    var idx;
    var items = _getSelectedTreeItems();
    for (var i=0; i < items.length; i++) {
        idx = gSCCItems.indexOf(items[i]);
        if (idx >= 0) {
            gSCCItems.splice(idx, 1);
            if (items[i] instanceof koSCCLoadingItem) {
                items[i].aborted = true;
            }
        }
    }
    gURLTreeView.setTreeItems(gSCCItems);
}

function CheckSelectedItems()
{
    _getSelectedTreeItems().forEach(function(elem) { elem.isChecked = true; });
    updateCommitButtonEnabledState();
    gURLTreeView.tree.invalidate();
}

function UncheckSelectedItems()
{
    _getSelectedTreeItems().forEach(function(elem) { elem.isChecked = false; });
    updateCommitButtonEnabledState();
    gURLTreeView.tree.invalidate();
}

function StopSelectedItems()
{
    var idx;
    var items = _getSelectedTreeItems();
    var changed_scc_listing = false;
    for (var i=0; i < items.length; i++) {
        if (items[i] instanceof koSCCLoadingItem) {
            items[i].aborted = true;
            idx = gSCCItems.indexOf(items[i]);
            if (idx >= 0) {
                gSCCItems.splice(idx, 1);
                changed_scc_listing = true;
            }
        }
    }
    if (changed_scc_listing) {
        gURLTreeView.setTreeItems(gSCCItems);
    }
}

function ContextPopupShowing(popupElement)
{
    var items = _getSelectedTreeItems();
    var stopMenu = document.getElementById("context_menu_stop");
    if (items.some(function(elem) { return (elem instanceof koSCCLoadingItem); })) {
        if (items.every(function(elem) { return (elem instanceof koSCCLoadingItem); })) {
            // All are loading items, display the stop menu as enabled.
            stopMenu.setAttribute("enabled", "true");
        } else {
            // Some items are loading items, display stop as grayed out.
            stopMenu.setAttribute("enabled", "false");
        }
        stopMenu.setAttribute("collapsed", "false");
    } else {
        // No loading items, hide the stop menu.
        stopMenu.setAttribute("collapsed", "true");
    }
}

// --- Button handlers.

var g_loaded_history = false;
var g_message_history = [];

function setMessageFromHistoryPosition(position) {
    document.getElementById("message").value = g_message_history[position];
}

function LoadCommitHistoryMenupopup(menupopup) {
    // We only need to load the history once.
    if (g_loaded_history) {
        return;
    }
    g_loaded_history = true;

    // Load the previous commit messages (from prefs) into the menupopup.
    var prefsSvc = Components.classes["@activestate.com/koPrefService;1"].
                        getService(Components.interfaces.koIPrefService);
    /**
     * @type {Components.interfaces.koIPreferenceSet}
     */
    var prefs = prefsSvc.prefs;
    var menuitem;
    if (prefs.hasPref("commit_message_history")) {
        /**
         * Ordered preference of commit messages (strings).
         * @type {Components.interfaces.koIOrderedPreference}
         */
        var orderedPref = prefs.getPref("commit_message_history");
        var message;
        for (var i=0; i < orderedPref.length; i++) {
            message = orderedPref.getStringPref(i);
            g_message_history.push(message);
            menuitem = document.createElement("menuitem");
            if (message.length <= 50) {
                menuitem.setAttribute("label", message);
            } else {
                menuitem.setAttribute("label", message.substr(0, 50) + "...");
                menuitem.setAttribute("tooltiptext", message);
            }
            menuitem.setAttribute("oncommand", "setMessageFromHistoryPosition(" + i + ");");
            menupopup.appendChild(menuitem);
        }
    }
    if (g_message_history.length == 0) {
        menuitem = document.createElement("menuitem");
        menuitem.setAttribute("label", _bundle.GetStringFromName("noMessageHistory"));
        menupopup.appendChild(menuitem);
    }
}

function AddFile()
{
    var basePath = gBaseUri;
    if (basePath) {
        basePath = ko.uriparse.URIToLocalPath(gBaseUri);
    }
    var paths = ko.filepicker.browseForFiles(basePath);
    if (paths) {
        // Turn into URIs
        var uris = [];
        var uri;
        for (var i=0; i < paths.length; i++) {
            uri = ko.uriparse.pathToURI(paths[i]);
            if (_urlIsInSccItemsList(uri)) {
                continue;
            }
            uris.push(uri);
        };

        var baseUri;
        if (gBaseUri) {
            baseUri = ko.uriparse.commonURIPrefixFromURIs(uris.concat([gBaseUri]));
        } else {
            baseUri = determineBaseUriFromUrls(uris);
        }
        if (baseUri != gBaseUri) {
            rebaseSccItemUris(baseUri);
        }
        addChosenSccFiles(uris);
        gURLTreeView.GetCommitableFiles(uris, false);
    }
}

function AddDirectory()
{
    var basePath = gBaseUri;
    if (basePath) {
        basePath = ko.uriparse.URIToLocalPath(gBaseUri);
    }
    var folder = ko.filepicker.getFolder(basePath);
    if (folder) {
        var uri = ko.uriparse.pathToURI(folder);
        var baseUri = uri;
        if (basePath) {
            baseUri = ko.uriparse.commonURIPrefixFromURIs([gBaseUri, uri]);
        }
        if (baseUri != gBaseUri) {
            //rebaseSccItemUris(baseUri);
            gBaseUri = baseUri;
        }

        var recursive = true;
        gURLTreeView.GetCommitableFiles([uri], recursive);
    }
    
}

function CheckAll()
{
    gURLTreeView.rows.forEach(function(elem) { elem.isChecked = true; });
    updateCommitButtonEnabledState();
    gURLTreeView.tree.invalidate();
}


function UncheckAll()
{
    gURLTreeView.rows.forEach(function(elem) { elem.isChecked = false; });
    updateCommitButtonEnabledState();
    gURLTreeView.tree.invalidate();
}

function ShowDiff()
{
    var relpaths = _getCheckedRelativePaths();
    if (relpaths.length > 0) {
        ko.scc.DiffRelative(gSCCSvc, gBaseUri, relpaths);
    }
}

function _abortAllLoadingItems()
{
    for (var scc_item in gSCCItems) {
        if (scc_item instanceof koSCCLoadingItem) {
            // XXX - We are only marking the item as aborted. The process will
            //       still continue anyway. It would be good to have a real way
            //       to stop this process.
            loadingItem.aborted = true;
        }
    }
}

function ConfirmCloseWindow()
{
    // If there is a commit message, then prompt to confirm the close action.
    var messageTextbox = document.getElementById("message");
    var message = messageTextbox.value.replace(/(^\s*|\s*$)/g, ''); // strip whitespace;
    if (message) {
        var prompt = "Are you sure you wish to cancel this commit. " +
                     "Your commit message will be lost.";
        var answer = ko.dialogs.yesNo(prompt, "Yes", null, null,
                                      "scc_no_commit_confirm_close");
        if (answer != "Yes") {
            return;
        }
    }
    // Close the window.
    Cancel();
}

function Cancel()
{
    _abortAllLoadingItems();
    window.arguments[0].retval = "Cancel";
    window.close();
}

function OK()
{
    var urls = _getCheckedUrls();
    if (!urls) {
        ko.dialogs.alert("No files have been checked.");
        return false;
    }

    _abortAllLoadingItems();

    var prefsSvc = Components.classes["@activestate.com/koPrefService;1"].
                        getService(Components.interfaces.koIPrefService);
    /**
     * @type {Components.interfaces.koIPreferenceSet}
     */
    var prefs = prefsSvc.prefs;

    // Validate the message, i.e. encourage (but don't enforce) the user actual
    // enter something for a message.
    var messageTextbox = document.getElementById("message");
    var message = messageTextbox.value;
    var stripped = message.replace(/(^\s*|\s*$)/g, ''); // strip whitespace
    if (!stripped) {
        var prompt = _bundle.GetStringFromName("noCommitMessage");
        if (prefs.getBooleanPref("donotask_scc_no_commit_message") &&
            prefs.getStringPref("donotask_action_scc_no_commit_message") == "No") {
            // A do not ask again of "No" does not make sense - as the commit
            // dialog will just flash briefly and the user will not know why they
            // cannot commit their empty message, so we reset it in this case.
            prefs.setBooleanPref("donotask_scc_no_commit_message", false);
            prefs.setStringPref("donotask_action_scc_no_commit_message", "");
        }
        var answer = ko.dialogs.yesNo(prompt, "No", null, null,
                                      "scc_no_commit_message");
        if (answer == "No") {
            messageTextbox.setSelectionRange(0, messageTextbox.textLength);
            messageTextbox.focus();
            return false;
        }
    }
    window.arguments[0].message = message;
    window.arguments[0].urls = urls;
    window.arguments[0].retval = "OK";
    if (gType == 'changelist') {
        var changelist_textbox = document.getElementById("changelist_name_textbox");
        window.arguments[0].changelistName = changelist_textbox.value;
    }

    // Remember the commit message.
    if (prefs.hasPref("commit_message_history")) {
        /**
         * Ordered preference of commit messages (strings).
         * @type {Components.interfaces.koIOrderedPreference}
         */
        var orderedPref = prefs.getPref("commit_message_history");
        // See if the message is already in the history, if so bump it's
        // position.
        for (var i=0; i < orderedPref.length; i++) {
            if (orderedPref.getStringPref(i) == message) {
                orderedPref.deletePref(i);
                break;
            }
        }
        orderedPref.insertStringPref(0, message);
        // Only keep the last 20 commit messages.
        while (orderedPref.length > 20) {
            orderedPref.deletePref(20);
        }
        // Must do this, otherwise the prefs are never saved - bug 83916.
        prefs.setPref("commit_message_history", orderedPref);
    }

    // Close the window and make the callback.
    window.close();
    window.arguments[0].callbackFn(window.arguments[0]);
    window.arguments[0].callbackFn = null;
    return true;
}
